/** GameBoy Cartridge Info and Manufacturer Codes ************/
/**                                                         **/
/**                        GBCarts.h                        **/
/**                                                         **/
/** This file contains functions to extract information     **/
/** from GameBoy cartridge images. It also includes a list  **/
/** of known manufacturer codes. Also see GBCarts.c.        **/
/**                                                         **/
/** Copyright (C) Marat Fayzullin 1995-1999                 **/
/**     You are not allowed to distribute this software     **/
/**     commercially. Please, notify me, if you make any    **/
/**     changes to this file.                               **/
/*************************************************************/
#ifndef GBCARTS_H
#define GBCARTS_H

#define GB_CRC(Header)       (((int)Header[0x14E]<<8)+Header[0x14F])
#define GB_ROMSize(Header)   (GB_ROMBanks(Header)<<14)
#define GB_CMP(Header)       Header[0x14D]
#define GB_Version(Header)   Header[0x14C]
#define GB_Type(Header)      Header[0x147]
#define GB_ValidType(Header) (GB_TypeName(Header)!=0)
#define GB_Developer(Header) (GB_CMP(Header)==0xE7)
#define GB_ColorGB(Header)   (Header[0x143]==0x80)
#define GB_SuperGB(Header)   (Header[0x146]==0x03)
#define GB_Japanese(Header)  (Header[0x14A]==0x00)
#define GB_Camera(Header)    (GB_Type(Header)==0xFC)
#define GB_Timer(Header)     ((GB_Type(Header)>=0x0F)&&(GB_Type(Header)<=0x10))
#define GB_Rumble(Header)    ((GB_Type(Header)>=0x1C)&&(GB_Type(Header)<=0x1E))
#define GB_MBC1(Header)      ((GB_Type(Header)>=0x01)&&(GB_Type(Header)<=0x03))
#define GB_MBC2(Header)      ((GB_Type(Header)>=0x05)&&(GB_Type(Header)<=0x06))
#define GB_MMM01(Header)     ((GB_Type(Header)>=0x0B)&&(GB_Type(Header)<=0x0D))
#define GB_MBC3(Header)      ((GB_Type(Header)>=0x0F)&&(GB_Type(Header)<=0x13))
#define GB_MBC4(Header)      ((GB_Type(Header)>=0x15)&&(GB_Type(Header)<=0x17))
#define GB_MBC5(Header)      ((GB_Type(Header)>=0x19)&&(GB_Type(Header)<=0x1E))
#define GB_HuC1(Header)      (GB_Type(Header)==0xFF)
#define GB_HuC3(Header)      (GB_Type(Header)==0xFE)

/** GB_MakerID() *********************************************/
/** Cartridge producer ID code.                             **/
/*************************************************************/
unsigned char GB_MakerID(unsigned char *Header);

/** GB_ROMBanks() ********************************************/
/** Size of on-cartridge ROM in 16kB pages.                 **/
/*************************************************************/
int GB_ROMBanks(unsigned char *Header);

/** GB_RAMSize() *********************************************/
/** Size of on-cartridge RAM in bytes.                      **/
/*************************************************************/
int GB_RAMSize(unsigned char *Header);

/** GB_Name() ************************************************/
/** Extract and truncate cartridge name. Returns a pointer  **/
/** to internal buffer!                                     **/
/*************************************************************/
char *GB_Name(unsigned char *Header);

/** GB_Maker() ***********************************************/
/** Try to figure out producer's name. Returns name or NULL **/
/** if no producer is known for cartridge's GB_MakerID().   **/
/*************************************************************/
char *GB_Maker(unsigned char *Header);

/** GB_TypeName() ********************************************/
/** Returns the textual name of the cartridge MBC type or   **/
/** NULL for unknown types.                                 **/
/*************************************************************/
char *GB_TypeName(unsigned char *Header);

/** GB_RealCMP() *********************************************/
/** Calculate complement byte of a cartridge.               **/
/*************************************************************/
unsigned char GB_RealCMP(unsigned char *Header);

/** GB_RealCRC() *********************************************/
/** Calculate CRC of a cartridge.                           **/
/*************************************************************/
unsigned short GB_RealCRC(unsigned char *Data);

/** GB_Battery() *********************************************/
/** Return 1 if cartridge has a battery, 0 otherwise.       **/
/*************************************************************/
unsigned char GB_Battery(unsigned char *Header);

#endif /* GBCARTS_H */
