/** EMULib Emulation Library *********************************/
/**                                                         **/
/**                         LibMSDOS.h                      **/
/**                                                         **/
/** This file contains definitions and declarations for     **/
/** routines in LibMSDOS.c.                                 **/
/**                                                         **/
/** Copyright (C) Marat Fayzullin 1996-1999                 **/
/**     You are not allowed to distribute this software     **/
/**     commercially. Please, notify me, if you make any    **/
/**     changes to this file.                               **/
/*************************************************************/
#ifndef LIBMSDOS_H
#define LIBMSDOS_H
#ifdef MSDOS

/*** Bitmap Images *******************************************/
typedef struct { char *Data;int W,H; } Image;
typedef unsigned char pixel;

void VGAPutImage(Image *Img,int X,int Y,int W,int H);
int  VGANewImage(Image *Img,int W,int H);
void VGAFreeImage(Image *Img);
/*************************************************************/

/*** Screen Modes ********************************************/
void VGA320x200(void);         /* Set 320x200x256 screenmode */
void VGAText80(void);          /* Set 80x25 text screenmode  */
void VGA256x240(void);         /* Set 256x240x256 screenmode */
void VGASetSync(int Switch);   /* Sync/don't sync to VBlanks */
/*************************************************************/

/*** Color Allocation ****************************************/
#define MAXALLOC  4096

pixel VGAGetColor(pixel R,pixel G,pixel B);
    /* Allocate a new color. Returns 0 (black color) on      */
    /* failure.                                              */
void VGAFlushColors(int N);
    /* Free all colors except for the N allocated last.      */
/*************************************************************/

/*** Joysticks ***********************************************/
typedef struct { int X[2],Y[2],Fire[2]; } JoyPos;

#define STICK1  0x01
#define STICK2  0x02

int GetJoyPos(JoyPos *Pos,int Sticks); /* Get jstk positions */
    /* Returns 0 or sticks for which it read positions.      */
/*************************************************************/

/*** Filename Processing *************************************/
char *GetPath(char *Name);     /* Get pathname from filename */
    /* Returns pointer to the internal buffer containing a   */
    /* path name ending with "\".                            */
char *NewFile(char *Name);     /* Generate a new filename    */
    /* Returns pointer to the internal buffer containing a   */
    /* file name in format NAMEnnnn.EXT (nnnn = 0000..9999). */
/*************************************************************/

/*** Saving/Loading Screen Snapshots *************************/
int LoadGIF(char *File);       /* Load screen from .GIF file */
    /* Returns number of allocated colors or 0 if failed.    */
int SaveGIF(char *File);       /* Save screen into .GIF file */
    /* Returns number of written scanlines or 0 if failed.   */
/*************************************************************/

#endif /* MSDOS */
#endif /* LIBMSDOS_H */