/** EMULib Emulation Library *********************************/
/**                                                         **/
/**                         SndMSDOS.c                      **/
/**                                                         **/
/** This file contains standard sound generation routines   **/
/** for MSDOS using SoundBlaster and Adlib.                 **/
/**                                                         **/
/** Copyright (C) Marat Fayzullin 1996-1999                 **/
/**     You are not allowed to distribute this software     **/
/**     commercially. Please, notify me, if you make any    **/
/**     changes to this file.                               **/
/*************************************************************/
#ifdef MSDOS

#include "SndMSDOS.h"

#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <dos.h>
#include <i86.h>

static unsigned short Freqs[] =
{
#include "FMFreq.h"
};

static unsigned char Volumes[256] =
{
  0xFF,0x3F,0x3F,0x3B,0x37,0x34,0x32,0x30,
  0x2E,0x2D,0x2B,0x2A,0x29,0x28,0x27,0x26,
  0x25,0x24,0x23,0x23,0x22,0x21,0x21,0x20,
  0x20,0x1F,0x1E,0x1E,0x1D,0x1D,0x1D,0x1C,
  0x1C,0x1B,0x1B,0x1A,0x1A,0x1A,0x19,0x19,
  0x19,0x18,0x18,0x18,0x17,0x17,0x17,0x17,
  0x16,0x16,0x16,0x15,0x15,0x15,0x15,0x14,
  0x14,0x14,0x14,0x14,0x13,0x13,0x13,0x13,
  0x12,0x12,0x12,0x12,0x12,0x11,0x11,0x11,
  0x11,0x11,0x10,0x10,0x10,0x10,0x10,0x10,
  0x0F,0x0F,0x0F,0x0F,0x0F,0x0F,0x0E,0x0E,
  0x0E,0x0E,0x0E,0x0E,0x0E,0x0D,0x0D,0x0D,
  0x0D,0x0D,0x0D,0x0D,0x0C,0x0C,0x0C,0x0C,
  0x0C,0x0C,0x0C,0x0C,0x0B,0x0B,0x0B,0x0B,
  0x0B,0x0B,0x0B,0x0B,0x0B,0x0A,0x0A,0x0A,
  0x0A,0x0A,0x0A,0x0A,0x0A,0x0A,0x09,0x09,
  0x09,0x09,0x09,0x09,0x09,0x09,0x09,0x08,
  0x08,0x08,0x08,0x08,0x08,0x08,0x08,0x08,
  0x08,0x08,0x07,0x07,0x07,0x07,0x07,0x07,
  0x07,0x07,0x07,0x07,0x07,0x06,0x06,0x06,
  0x06,0x06,0x06,0x06,0x06,0x06,0x06,0x06,
  0x06,0x05,0x05,0x05,0x05,0x05,0x05,0x05,
  0x05,0x05,0x05,0x05,0x05,0x05,0x04,0x04,
  0x04,0x04,0x04,0x04,0x04,0x04,0x04,0x04,
  0x04,0x04,0x04,0x04,0x04,0x03,0x03,0x03,
  0x03,0x03,0x03,0x03,0x03,0x03,0x03,0x03,
  0x03,0x03,0x03,0x03,0x02,0x02,0x02,0x02,
  0x02,0x02,0x02,0x02,0x02,0x02,0x02,0x02,
  0x02,0x02,0x02,0x02,0x01,0x01,0x01,0x01,
  0x01,0x01,0x01,0x01,0x01,0x01,0x01,0x01,
  0x01,0x01,0x01,0x01,0x01,0x01,0x00,0x00,
  0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00
};

static unsigned short OpAd[] =
{
  0x000,0x001,0x002,0x008,0x009,0x00A,0x010,0x011,0x012,
  0x100,0x101,0x102,0x108,0x109,0x10A,0x110,0x111,0x112
};

static unsigned short ChanAd[] =
{
  0x000,0x001,0x002,0x003,0x004,0x005,0x006,0x007,0x008,
  0x100,0x101,0x102,0x103,0x104,0x105,0x106,0x107,0x108
};

static int SndRate=0;    /* Sound rate (1=Adlib, 0=Off)      */
static int XFreq[7];     /* Internal storage for frequencies */
static int XVolume[7];   /* Internal storage for volumes     */
static int XType[7];     /* Sound types                      */
static int MasterSwitch; /* Switches to turn channels on/off */
static int MasterVolume; /* Master volume                    */
static int MaxFreq;      /* Maximal possible frequency       */
static int SndLock;      /* Internal lock                    */
static int SndPending;   /* SetChannels() call is pending    */

static int SBPort,SBIRQ,SBDMA;
static int NoiseGen,CurBuffer;
static unsigned char *Buffer;
static unsigned short Segment=0;
static int Wave[SND_BUFSIZE];
static int Count[SND_CHANNELS];
static void (interrupt *IRQOld)()=0;
static void interrupt IRQHandler(void);

static void AdlibOut(int R,int V); /* Write value into Adlib */
static int ReadDSP(void);          /* Read value from DSP    */
static int WriteDSP(int V);        /* Write value into DSP   */
static void WriteMixer(int R,int V); /* Write V into mixer   */
static void Lock(void);            /* Set internal lock      */
static void Unlock(void);          /* Release internal lock  */

/** InitSound() **********************************************/
/** Initialize sound. Returns Rate on success, 0 otherwise. **/
/** Rate=0 to skip initialization (will be silent), Rate=1  **/
/** to use Adlib, Rate=8192..44100 to use SB wave synthesis.**/
/*************************************************************/
int InitSound(int Rate)
{
  int I;

  /* Trash sound if it was initialized */
  TrashSound();

  /* Initialize internal variables */
  MaxFreq=0;SndRate=0;SndLock=0;
  SndPending=0;
  Buffer=0;Segment=0;IRQOld=0;
  MasterVolume=0;MasterSwitch=0x00;
  NoiseGen=1;

  /* Reset sound parameters */
  for(I=0;I<7;I++)
  {
    Count[I]=XVolume[I]=XFreq[I]=0;
    XType[I]=SND_MELODIC;
  }

  /* Initializing the right hardware */
  switch(Rate)
  {
    case 0:
      return(1);
    case 1:
      if(!InitAdlib()) return(0);
      MaxFreq=sizeof(Freqs)*5;
      break;
    default:
      Rate=InitSB(Rate);
      if(!Rate) return(0);
      MaxFreq=Rate/2;
      break;
  }

  SetChannels(192,0xFF);
  return(SndRate=Rate);
}

/** TrashSound() *********************************************/
/** Shut down sound.                                        **/
/*************************************************************/
void TrashSound(void)
{
  switch(SndRate)
  {
    case 0:
      break;
    case 1:
      /* Adlib */
      SetChannels(0,0x00);    /* Turn off the sound      */
      AdlibOut(0x04,0x60);    /* Mask and disable timers */
      AdlibOut(0x04,0x80);    /* Reset timers            */
      break;
    default:
      /* SoundBlaster */
      WriteDSP(0xD3);WriteDSP(0xD0);   /* Reset DSP      */
      outp(0x0A,SBDMA|0x04);           /* Disable DMA    */
      outp(0x21,inp(0x21)|(1<<SBIRQ)); /* Disable IRQ    */
      break;
  }

  /* If SB interrupt vector changed, restore it */
  if(IRQOld) { _dos_setvect(SBIRQ+8,IRQOld);IRQOld=0; }
  /* If low memory was allocated for DMA, free it */
  if(Segment) { _dos_freemem(Segment);Segment=0; }
  /* Sound trashed */
  SndRate=0;
}

/** SetSound() ***********************************************/
/** Does nothing now. Supposed to set sound type (melodic/  **/
/** noise) for a channel.                                   **/
/*************************************************************/
void SetSound(int Channel,int Type)
{
  if((SndRate==1)&&(XType[Channel]!=Type))
  {
    Lock();
    AdlibOut(0x43+OpAd[XType[Channel]==SND_NOISE? 7:Channel],0xFF);
    AdlibOut(0x43+OpAd[Type==SND_NOISE? 7:Channel],0xFF);
    XFreq[Channel]=XVolume[Channel]=0;
    Unlock();
  }
  XType[Channel]=Type;
}

/** Sound() **************************************************/
/** Generate sound of given frequency (Hz) and volume       **/
/** (0..255) via given channel.                             **/
/*************************************************************/
void Sound(register int Channel,register int Freq,register int Volume)
{
  register int N;

  if(Freq>=MaxFreq) Freq=0;
  if(!Freq) Volume=0;
  Volume&=0xFF;
  XVolume[Channel]=Volume;
  XFreq[Channel]=Freq;

  if(SndRate==1)
  {
    Lock();
    /* Noise is played via separate channel */
    if(XType[Channel]==SND_NOISE) { N=7;Volume/=2; } else N=Channel;
    /* Volume */
    if(MasterSwitch&(1<<Channel))
      AdlibOut(0x43+OpAd[N],Volumes[Volume*MasterVolume/256]);
    /* Frequency */
    if(XType[Channel]!=SND_NOISE)
    {
      Freq/=5;
      AdlibOut(0xA0+ChanAd[N],Freqs[Freq]&0xFF);
      AdlibOut(0xB0+ChanAd[N],Freqs[Freq]>>8);
    }
    Unlock();
  }
}

/** SetFreq() ************************************************/
/** Set sound frequency (Hz) of a channel                   **/
/*************************************************************/
void SetFreq(register int Channel,register int Freq)
{
  if(Freq>=MaxFreq) Freq=0;
  if(!Freq) SetVolume(Channel,0);
  XFreq[Channel]=Freq;

  if((SndRate==1)&&(XType[Channel]!=SND_NOISE))
  {
    Lock();
    Freq/=5;
    AdlibOut(0xA0+ChanAd[Channel],Freqs[Freq]&0xFF);
    AdlibOut(0xB0+ChanAd[Channel],Freqs[Freq]>>8);
    Unlock();
  }
}

/** SetVolume() **********************************************/
/** Set sound volume (0..255) of a channel.                 **/
/*************************************************************/
void SetVolume(register int Channel,register int Volume)
{
  Volume&=0xFF;
  XVolume[Channel]=Volume;

  if(SndRate==1)
  {
    Lock();
    if(MasterSwitch&(1<<Channel))
    {
      if(XType[Channel]==SND_NOISE) { Channel=7;Volume/=2; }
      AdlibOut(0x43+OpAd[Channel],Volumes[Volume*MasterVolume/256]);
    }
    Unlock();
  }
}

/** SetChannels() ********************************************/
/** Set master volume (0..255) and turn channels on/off.    **/
/** Each bit in Toggle corresponds to a channel (1=on).     **/
/*************************************************************/
void SetChannels(int Volume,int Toggle)
{
  register int I;

  Volume&=0xFF;
  MasterVolume=Volume;
  MasterSwitch=Toggle;

  /* When using DSP, change master volume via SB mixer */
  if(SndRate>=8192) WriteMixer(0x22,(Volume&0xF0)|(Volume>>4));

  if(SndRate==1)
  {
    /* Set pending flag if sound is locked */
    if(SndLock) { SndPending=1;return; } else SndLock=1;

    for(I=0;I<7;I++)
      if(XType[I]==SND_NOISE)
        AdlibOut
        (0x43+OpAd[7],Toggle&(1<<I)? Volumes[XVolume[I]*Volume/512]:0xFF);
      else
        AdlibOut
        (0x43+OpAd[I],Toggle&(1<<I)? Volumes[XVolume[I]*Volume/256]:0xFF);

    Unlock();
  }
}

/** Lock()/Unlock() ******************************************/
/** Set/Release internal lock.                              **/
/*************************************************************/
void Lock(void) { while(SndLock) {};SndLock++; }
void Unlock(void)
{
  if(!--SndLock)
    if(SndPending)
    { SndPending=0;SetChannels(MasterVolume,MasterSwitch); }
}

/** AdlibOut() ***********************************************/
/** Write value V into Adlib register R.                    **/
/*************************************************************/
void AdlibOut(register int R,register int V)
{
  register int J;
  outp(0x388,R);for(J=0;J<6;J++) inp(0x388);
  outp(0x389,V);for(J=0;J<35;J++) inp(0x388);
}

/** ReadDSP() ************************************************/
/** Read a value from SoundBlaster. Returns -1 if failed.   **/
/*************************************************************/
int ReadDSP(void)
{
  register int J;

  for(J=0;J<100000;J++)
    if(inp(SBPort+0x0E)&0x80)
      return(inp(SBPort+0x0A));
  return(-1);
}

/** WriteDSP() ***********************************************/
/** Write a value into SoundBlaster. Returns 0 if failed,   **/
/** 1 otherwise.                                            **/
/*************************************************************/
int WriteDSP(register int V)
{
  register int J;

  for(J=0;J<100000;J++)
    if(!(inp(SBPort+0x0C)&0x80))
    { outp(SBPort+0x0C,V);return(1); }
  return(0);
}

/** WriteMixer() *********************************************/
/** Write a value into SoundBlaster's mixer.                **/
/*************************************************************/
void WriteMixer(register int R,register int V)
{ outp(SBPort+0x04,R);outp(SBPort+0x05,V); }

/** InitAdlib() **********************************************/
/** Initialize Adlib.                                       **/
/*************************************************************/
int InitAdlib(void)
{
  int I,A1,A2;

  /* No interrupts for now */
  _disable();

  /* Detect if Adlib is present */
  AdlibOut(0x04,0x60);    /* Mask and disable timers */
  AdlibOut(0x04,0x80);    /* Reset timers            */
  A1=inp(0x388)&0xE0;     /* Read status             */
  AdlibOut(0x02,0xFF);    /* Set Timer1 to 0xFF      */
  AdlibOut(0x04,0x21);    /* Unmask and start Timer1 */
  for(I=0;I<0xC8;I++) inp(0x388);  /* Wait 80us      */
  A2=inp(0x388)&0xE0;     /* Read status             */
  AdlibOut(0x04,0x60);    /* Mask and disable timers */
  AdlibOut(0x04,0x80);    /* Reset timers            */
  if((A2!=0xC0)||A1) { _enable();return(0); }

  /* Allow the FM chips to control waveform */
  AdlibOut(0x01,0x20);
  AdlibOut(0xBD,0x20);

  /* Turn all channels off */
  for(I=0;I<8;I++) AdlibOut(0x43+OpAd[I],0xFF);

  for(I=0;I<3;I++)
  {
    AdlibOut(0x20+I,0x01);
    AdlibOut(0x23+I,0x01);
    AdlibOut(0x40+I,0x18);
    AdlibOut(0x43+I,0x3F);
    AdlibOut(0x60+I,0xF0);
    AdlibOut(0x63+I,0xF0);
    AdlibOut(0x80+I,0x14);
    AdlibOut(0x83+I,0x13);
    AdlibOut(0xE0+I,0x02);
    AdlibOut(0xE3+I,0x00);
    AdlibOut(0xC0+I,0x0A);
    AdlibOut(0xB0+I,0x00);
  }
  AdlibOut(0x34,0x21);
  AdlibOut(0x54,0x3F);
  AdlibOut(0x74,0x99);
  AdlibOut(0x94,0x00);
  AdlibOut(0xF4,0x00);
  AdlibOut(0xBD,0x28);

  /* Set fixed frequency for noise */
  AdlibOut(0xA0+ChanAd[7],0x9F);
  AdlibOut(0xB0+ChanAd[7],0x3E);

  /* Back to business */
  _enable();
  return(1);
}

/** InitSB() *************************************************/
/** Initialize SoundBlaster for wave synthesis via DMA.     **/
/*************************************************************/
int InitSB(int Rate)
{
  static int PagePort[8] = { 0x87,0x83,0x81,0x82,-1,0x8B,0x89,0x8A };
  char *P;
  int J;

  /* If Rate is out of range, fall out */
  if((Rate<8192)||(Rate>44100)) Rate=22050;

  /* Default values */
  SBPort=0x220;SBIRQ=7;SBDMA=1;

  /* Get SoundBlaster Port#,IRQ#,DMA# */
  if(P=getenv("BLASTER"))
    while(*P)
      switch(toupper(*P++))
      {
        case 'A': SBPort=strtol(P,0,16);break;
        case 'D': SBDMA=strtol(P,0,16);break;
        case 'I': SBIRQ=strtol(P,0,16);break;
      }

  /* IRQ# = 0,1,2,3,4,5,6,7, DMA# = 0,1,2,3 */
  if((SBIRQ>7)||(SBDMA>3)) return(0);

  /* Reset DSP */
  outp(SBPort+0x06,0x01); /* Set the reset flag   */
  delay(100);             /* Wait for 100ms       */
  outp(SBPort+0x06,0x00); /* Clear the reset flag */
  /* Wait for READY status */
  for(J=100;J&&(ReadDSP()==0xAA);J--);
  if(!J) return(0);

  /* Allocate low memory for DMA buffer */
  J=SND_BUFSIZE*4/16+4;
  if(_dos_allocmem(J,&Segment)) { Segment=0;return(0); }
  J=(int)Segment*16;
  Buffer=(unsigned char *)J;
  if((J>>16)!=((J+2*SND_BUFSIZE-1)>>16)) Buffer+=2*SND_BUFSIZE;
  memset(Buffer,0,2*SND_BUFSIZE);

  /* Save old interrupt vector */
  IRQOld=(void (interrupt *)())_dos_getvect(SBIRQ+8);
  /* Install new interrupt vector */
  _dos_setvect(SBIRQ+8,IRQHandler);
  /* Enable the SoundBlaster IRQ */
  outp(0x21,inp(0x21)&~(1<<SBIRQ));

  /* Set up DMA to do the transfer */
  WriteDSP(0x40);
  WriteDSP((65536L-(256000000L/Rate))/256);
  _disable();
  outp(0x0A,SBDMA|0x04);
  outp(0x0C,0x00);
  outp(0x0B,SBDMA|0x58);
  J=SBDMA*2;
  outp(J,(int)Buffer&0xFF);
  outp(J,((int)Buffer>>8)&0xFF);
  J++;
  outp(J,(SND_BUFSIZE*2-1)&0xFF);
  outp(J,(SND_BUFSIZE*2-1)>>8);
  outp(PagePort[SBDMA],((int)Buffer>>16)&0xFF);
  outp(0x0A,SBDMA);
  _enable();
  WriteDSP(0xD1);
  WriteDSP(0x48);
  WriteDSP((SND_BUFSIZE-1)&0xFF);
  WriteDSP((SND_BUFSIZE-1)>>8);
  WriteDSP(0x90);
  CurBuffer=0;

  return(Rate);
}

/** IRQHandler() *********************************************/
/** This function handles SoundBlaster interrupts when DMA  **/
/** gets complete.                                          **/
/*************************************************************/
void interrupt IRQHandler(void)
{
  register unsigned char *P;
  register int I,J,K,L1,L2,V;
  static int Busy=0;

  /* Acknowledge interrupt */
  inp(SBPort+0x0E);
  outp(0x20,0x20);
  if(SBIRQ>=8) outp(0xA0,0x20);

  /* Fall out if recursive interrupt */
  if(Busy) { CurBuffer^=1;return; }
  Busy=1;_enable();

  /* Waveform generator */
  for(J=0;J<SND_CHANNELS;J++)
    if(XFreq[J]&&(V=XVolume[J])&&(MasterSwitch&(1<<J)))
      switch(XType[J])
      {
        case SND_MELODIC: /* Melodic Sound */
          K=0x10000*XFreq[J]/SndRate;
          L1=Count[J];
          for(I=0;I<SND_BUFSIZE;I++)
          {
            L2=L1+K;
            Wave[I]+=
              V*(L1&0x8000? (L2&0x8000? 32767:0):(L2&0x8000? 0:-32767));
            L1=L2;
          }
          Count[J]=L1;
          break;

        case SND_NOISE: /* White Noise */
          K=0x10000*XFreq[J]/SndRate;
          L1=Count[J];
          for(I=0;I<SND_BUFSIZE;I++)
          {
            L1+=K;
            if(L1&0xFFFF0000)
            {
              L1&=0xFFFF;
              if((NoiseGen<<=1)&0x80000000) NoiseGen^=0x08000001;
            }
            Wave[I]+=V*(NoiseGen&1? 32767:-32767);
          }
          Count[J]=L1;
          break;
      }

  /* Mix and convert waveforms */
  P=Buffer+CurBuffer*SND_BUFSIZE;
  for(J=0;J<SND_BUFSIZE;J++)
  { *P++=(Wave[J]/SND_CHANNELS)/256/512+128;Wave[J]=0; }
  CurBuffer^=1;
  Busy=0;
}

#endif /* MSDOS */