/** EMULib Emulation Library *********************************/
/**                                                         **/
/**                         SndWin.h                        **/
/**                                                         **/
/** This file contains standard sound generation routines   **/
/** for Windows using waveOut and midiOut.                  **/
/**                                                         **/
/** Copyright (C) Marat Fayzullin 1996-1999                 **/
/**     You are not allowed to distribute this software     **/
/**     commercially. Please, notify me, if you make any    **/
/**     changes to this file.                               **/
/*************************************************************/
#ifndef SNDWIN_H
#define SNDWIN_H
#ifdef WINDOWS

#include <Windows.h>

#define SND_MELODIC  0x01      /* Melodic sound (default)    */
#define SND_NOISE    0x02      /* White noise                */
#define SND_PERIODIC 0x04      /* Periodic noise (not im-ed) */

#define SND_CHANNELS    5      /* Number of channels         */
#define SND_BUFSIZE   256      /* Size of a wave buffer      */
#define SND_BUFFERS    32      /* Number of wave buffers     */

int InitSound(HINSTANCE hInst,int Rate,int Buffers);
    /* Initialize sound with given synthesis rate and number */
    /* of buffers. Returns 0 if failed. Use Rate=0 to skip   */
    /* initialization (will be silent), or synthesis rate    */
    /* (8192-44100Hz) otherwise. Values of 1-8191 will make  */
    /* it use MIDI mapper to play music. Number of buffers   */
    /* should be in 2..SND_BUFFERS range.                    */

void TrashSound(void);                    /* Shut down sound */
void SetSound(int Channel,int NewType);   /* Set sound type  */
void SetFreq(int Channel,int NewFreq);    /* Set frequency   */
void SetVolume(int Channel,int NewVolume);/* Set volume      */

void SetChannels(int Volume,int Switch);
    /* Set master volume (0..255) and turn channels on/off.  */
    /* Each bit in Switch corresponds to a channel (1=on).   */

void Sound(int Channel,int NewFreq,int NewVolume);
    /* Generate sound of given frequency (Hz) and volume     */
    /* (0..255) via given channel.                           */

#endif /* WINDOWS */
#endif /* SNDWIN_H */